/*
 *  bfx.c: FX network filter driver
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/errno.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/if_arp.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/version.h>
#include <linux/device.h>

#include <linux/if_bridge.h>
#include <linux/netfilter_bridge.h>
#include "br_private.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#  error This driver support 2.6.x kernels only!
#endif

/* devfs_register_chrdev calls register_chrdev and if you pass 0
 * as the major device number it dynamically allocates one.
 */

/* We are going to use the major number of the BDM commercial debugger
 * until we can obtain our own major number or devfs becomes the standard.
 */
#define FX_MAJOR 245

static int fx_major = -1;
static struct proc_dir_entry *fx_ent = NULL;
struct class_simple *fx_class = NULL;

#define FX_MAX_INSTANCE 20

typedef struct _fx_buf {
	struct net_device *dev;
	struct sk_buff_head inbuf;
	wait_queue_head_t inwait;
	char in_opened;
	struct sk_buff_head outbuf;
	wait_queue_head_t outwait;
	char out_opened;
	spinlock_t race_lock;
} FxBuf;

FxBuf *fxroot[FX_MAX_INSTANCE];
int pleaseexit[FX_MAX_INSTANCE];
char EnablementCode[] = "FXENABLED1";

int query_bridge_for = -1;


unsigned int hook_in(unsigned int hooknum,
	struct sk_buff **skb,
	const struct net_device *in,
	const struct net_device *out,
	int (*okfn)(struct sk_buff *));

unsigned int hook_out(unsigned int hooknum,
	struct sk_buff **skb,
	const struct net_device *in,
	const struct net_device *out,
	int (*okfn)(struct sk_buff *));

static struct
nf_hook_ops nfho_in = {
	hook:		hook_in,
	hooknum:	NF_BR_PRE_ROUTING,
	pf:		PF_BRIDGE,
	priority:	NF_BR_PRI_FIRST
};

static struct
nf_hook_ops nfho_out = {
	hook:		hook_out,
	hooknum:	NF_BR_POST_ROUTING,
	pf:		PF_BRIDGE,
	priority:	NF_BR_PRI_FIRST
};

static int fx_notifier_event(struct notifier_block *this,
	unsigned long event,
	void *ptr);

static struct notifier_block fx_dev_notifier = {
	.notifier_call	= fx_notifier_event
};

/* An always useful macro */
#define MIN(a, b) (((a) < (b)) ? (a) : (b))

#pragma pack(1)
struct udphdr {
	unsigned short uh_sport;	/* source port */
	unsigned short uh_dport;	/* destination port */
	short uh_ulen;		/* udp length */
	unsigned short uh_sum;	/* udp checksum */
};

struct ip {
	unsigned char ip_hl:4;	/* header length */
	unsigned char ip_v:4;	/* version */
	unsigned char ip_tos;	/* type of service */
	unsigned short ip_len;	/* total length */
	unsigned short ip_id;	/* identification */
	unsigned short ip_off;	/* fragment offset field */
	unsigned char ip_ttl;	/* time to live */
	unsigned char ip_p;	/* protocol */
	unsigned short ip_sum;	/* checksum */
	unsigned long ip_src, ip_dst;	/* source and dest address */
};
#pragma pack()



int IsDHCP(unsigned char *buffer)
{
	/* IP Packet */
	if (buffer[12] == 0x08 && buffer[13] == 0x00)
	{
		struct ip *pip = (struct ip *)&buffer[14];

		if (pip->ip_p == 17)
		{
			struct udphdr *pudp =
			    (struct udphdr *)&buffer[14 + (pip->ip_hl * 4)];
			unsigned short dp = ntohs(pudp->uh_dport);
			unsigned short sp = ntohs(pudp->uh_sport);

			if (dp == 67 || sp == 67 || dp == 68 || sp == 68)
				return 1;
		}
	}
	return 0;
}

int is_dhcp(struct sk_buff *skb)
{
	static unsigned char buf[114];

	if (skb->mac.raw)
	{
		int len = MIN(skb->len + ETH_HLEN, 100);

		memcpy(buf, skb->mac.raw, ETH_HLEN);

		memcpy(buf + ETH_HLEN, skb->data, len);
		return IsDHCP(buf);
	}
	return 0;
}



/* extract first 2 characters */
static int _atoi(const char *str)
{
	int ret = -1;
	char f, s;

	f = str[0];
	s = str[1];

	if(f >= '0' && f <= '9')
	{
		ret = f - '0';

		if(s >= '0' && s <= '9')
			ret = ret*10 + s - '0';
	}

	return ret;
}



static int _find_file_instance(struct file *file)
{
	char *name;

	if (!file || !file->f_dentry)
		return -1;

	name = (char *)file->f_dentry->d_name.name;

	if(name[2] == 'i')
		return _atoi(name+4); /* fxinXX */
	else
		return _atoi(name+5); /* fxoutXX */
}



static void free_fx_instance(int index)
{
	if (fxroot[index])
	{
		devfs_remove("fx/fxin%d", index);
		devfs_remove("fx/fxout%d", index);
		class_simple_device_remove(MKDEV(fx_major, (index * 2))); 
		class_simple_device_remove(MKDEV(fx_major, ((index * 2) + 1))); 

		skb_queue_purge(&fxroot[index]->inbuf);
		skb_queue_purge(&fxroot[index]->outbuf);

		kfree(fxroot[index]);
		fxroot[index] = NULL;
	}
}



static int init_fx_instance(int index, struct net_device *dev)
{
	if(fxroot[index])
	{
		printk(KERN_ERR "fx%d is already registered on eth%d",
			index, index);
		return -1;
	}

	/* Allocate a buffer for the current instance */
	fxroot[index] = kmalloc(sizeof(FxBuf), GFP_KERNEL);
	if(fxroot[index] == NULL)
	{
		printk(KERN_ERR "fx: not enough memory");
		return -1;
	}

	memset(fxroot[index], 0, sizeof(FxBuf));

	init_waitqueue_head(&(fxroot[index]->outwait));
	skb_queue_head_init(&(fxroot[index]->outbuf));
	init_waitqueue_head(&(fxroot[index]->inwait));
	skb_queue_head_init(&(fxroot[index]->inbuf));

	class_simple_device_add(fx_class, MKDEV(fx_major, (index * 2)),
				NULL, "fx/fxin%d", index);

	devfs_mk_cdev(MKDEV(fx_major, (index * 2)),
		      S_IFCHR | S_IRUSR | S_IWUSR, "fx/fxin%d",
		      index);

	class_simple_device_add(fx_class, MKDEV(fx_major, ((index * 2) + 1)),
				NULL, "fx/fxout%d", index);

	devfs_mk_cdev(MKDEV(fx_major, ((index * 2) + 1)),
		      S_IFCHR | S_IRUSR | S_IWUSR, "fx/fxout%d", index);

	fxroot[index]->dev = dev;

	return 0;
}



/* Routines to deal with /proc */
static int fx_proc_read(char *buffer, char **start, off_t offset,
			int length, int *eof, void *data)
{
	struct net_bridge_port *port;
	char *name;

	*start = buffer;

	/* Signal that this is the end
	 * of the file */
	if (offset > 0)
		return 0;

	if (length > 0 && query_bridge_for >= 0 && fxroot[query_bridge_for])
	{
		/* request for parent bridge brX index */
		port = rcu_dereference(fxroot[query_bridge_for]->dev->br_port);
		name = port->br->dev->name;
		if(name[3] > '0' && name[2] < '9')
		{
			buffer[0] = name[2];
			buffer[1] = name[3];
		}
		else
		{
			buffer[0] = '0';
			buffer[1] = name[2];
		}
		query_bridge_for = -1;

		buffer[2] = 'b';

		return 3;
	}

	return 0;
}

static int fx_proc_write(struct file *file, const char *buffer,
			 unsigned long count, void *data)
{
	int index;

	if(count == 3 && buffer[2] == 'b')
	{
		/* This limits us to us only 10 ifaces */
		index = _atoi(buffer);

		if (	index > -1 &&
			index < FX_MAX_INSTANCE &&
			fxroot[index])
		{
			/* request for parent bridge */
			if(fxroot[index]->dev->br_port)
				query_bridge_for = index;
		}
	}

	return count;
}


/* Dealing with the in (from ethernet) device */
static unsigned int fx_in_dev_poll(struct file *file, poll_table * wait)
{
	int instance = _find_file_instance(file);

	if (	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL ||
		fxroot[instance]->dev->br_port == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return 0;
	}

	if (skb_queue_len(&(fxroot[instance]->inbuf)))
		return POLLIN | POLLRDNORM;

	poll_wait(file, &(fxroot[instance]->inwait), wait);

	if (skb_queue_len(&(fxroot[instance]->inbuf)))
		return POLLIN | POLLRDNORM;

	return 0;
}


static ssize_t fx_in_dev_write(struct file *file, const char *buf,
			       size_t count, loff_t * pos)
{

	struct sk_buff *skb;
	int instance = _find_file_instance(file);

	if (	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL ||
		fxroot[instance]->dev->br_port == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return -1;
	}

	skb = dev_alloc_skb(count);
	if(skb)
	{
		skb->mac.raw = skb_put(skb, count);
		skb->nh.raw = skb->mac.raw + ETH_HLEN;

		copy_from_user(skb->mac.raw, buf, count);
		copy_from_user(&skb->protocol, buf + 2*ETH_ALEN, 2);

		skb->dev = fxroot[instance]->dev;
		skb->data = skb->mac.raw;

		dev_queue_xmit(skb);
	}

	return 0;
}

static ssize_t fx_in_dev_read(struct file *file, char *buf,
			      size_t count, loff_t * pos)
{
	int instance = _find_file_instance(file);
	struct sk_buff *skb;

	if(	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return -1;
	}

	skb = skb_dequeue(&(fxroot[instance]->inbuf));
	if(skb)
	{
		copy_to_user(buf, skb->mac.raw, skb->len + ETH_HLEN);

		kfree_skb(skb);

		return skb->len + ETH_HLEN;
	}

	return 0;
}

static loff_t fx_in_dev_lseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static int fx_in_dev_open(struct inode *inode, struct file *file)
{
	int instance = _find_file_instance(file);

	if(	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL ||
		fxroot[instance]->dev->br_port == NULL)
	{
		return -1;
	}

	fxroot[instance]->in_opened = 1;

	return 0;
}

static int fx_in_dev_release(struct inode *inode, struct file *file)
{
	int instance = _find_file_instance(file);

	if(	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return -1;
	}

	fxroot[instance]->in_opened = 0;


	return 0;
}

static int fx_in_dev_ioctl(struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg)
{
	/* We don't have any FX specific ioctls at the moment */
	return 0;
}

static struct file_operations fx_in_dev_fops = {
	owner:		THIS_MODULE,
	llseek:		fx_in_dev_lseek,
	read:		fx_in_dev_read,
	write:		fx_in_dev_write,
	poll:		fx_in_dev_poll,
	ioctl:		fx_in_dev_ioctl,
	open:		fx_in_dev_open,
	release:	fx_in_dev_release
};

/* Dealing with the out (from kernel TCP/IP) device */
static unsigned int fx_out_dev_poll(struct file *file, poll_table * wait)
{
	int instance = _find_file_instance(file);

	if (	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL ||
		fxroot[instance]->dev->br_port == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return 0;
	}

	if (skb_queue_len(&(fxroot[instance]->outbuf)))
		return POLLIN | POLLRDNORM;

	poll_wait(file, &(fxroot[instance]->outwait), wait);

	if (skb_queue_len(&(fxroot[instance]->outbuf)))
		return POLLIN | POLLRDNORM;

	return 0;
}

static ssize_t fx_out_dev_write(struct file *file, const char *buf,
				size_t count, loff_t * pos)
{
	struct sk_buff *skb;
	int instance = _find_file_instance(file);

	if (	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL ||
		fxroot[instance]->dev->br_port == NULL ||
		count < 15)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return -1;
	}

	skb = dev_alloc_skb(count);
	if (skb)
	{
		skb->mac.raw = skb_put(skb, count);

		copy_from_user(skb->mac.raw, buf, count);
		copy_from_user(&skb->protocol, buf + 2*ETH_ALEN, 2);

		skb->nh.raw = skb->mac.raw + ETH_HLEN;
		skb->data = skb->nh.raw;

		skb->dev = fxroot[instance]->dev;

		skb->cb[0] = 1;
		netif_rx(skb);

		return count;
	}

	return 0;
}

static ssize_t fx_out_dev_read(struct file *file, char *buf,
			       size_t count, loff_t * pos)
{
	int instance = _find_file_instance(file);
	struct sk_buff *skb;

	if (	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return -1;
	}

	skb = skb_dequeue(&(fxroot[instance]->outbuf));
	if(skb)
	{
		copy_to_user(buf, skb->mac.raw, skb->len + ETH_HLEN);

		kfree_skb(skb);

		return skb->len + ETH_HLEN;
	}

	return 0;
}

static loff_t fx_out_dev_lseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static int fx_out_dev_open(struct inode *inode, struct file *file)
{
	int instance = _find_file_instance(file);

	if(	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL ||
		fxroot[instance]->dev->br_port == NULL)
	{
		return -1;
	}

	fxroot[instance]->out_opened = 1;

	return 0;
}

static int fx_out_dev_release(struct inode *inode, struct file *file)
{
	int instance = _find_file_instance(file);

	if(	instance < 0 ||
		instance >= FX_MAX_INSTANCE ||
		fxroot[instance] == NULL)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
		return -1;
	}

	fxroot[instance]->out_opened = 0;

	return 0;
}

static int fx_out_dev_ioctl(struct inode *inode, struct file *file,
			    unsigned int cmd, unsigned long arg)
{
	/* We don't have any FX specific ioctls at the moment */
	return 0;
}

static struct file_operations fx_out_dev_fops = {
	owner:		THIS_MODULE,
	llseek:		fx_out_dev_lseek,
	read:		fx_out_dev_read,
	write:		fx_out_dev_write,
	poll:		fx_out_dev_poll,
	ioctl:		fx_out_dev_ioctl,
	open:		fx_out_dev_open,
	release:	fx_out_dev_release
};

/* Generic */
static int fx_generic_open(struct inode *inode, struct file *file)
{
	int len = 0;
	char *name;

	if (!file || !file->f_dentry)
		return -ENODEV;

	name = (char *)file->f_dentry->d_name.name;
	len = file->f_dentry->d_name.len;

	/* On a non-devfs system we have to set the file_operation
	 * structs ourselves since devfs_register() doesn't.
	 */
	if (len > 3 && strncmp("fxin", name, 4) == 0)
	{
		file->f_op = &fx_in_dev_fops;
		return fx_in_dev_open(inode, file);
	}
	else if (len > 4 && strncmp("fxout", name, 5) == 0)
	{
		file->f_op = &fx_out_dev_fops;
		return fx_out_dev_open(inode, file);
	}
	return -ENODEV;
}

static struct file_operations fx_generic_fops = {
	owner:		THIS_MODULE,
	open:		fx_generic_open
};


int fx_init_module(void)
{
	int ret;
	int err;

	printk(KERN_INFO "F/X Communications Network Filter Driver, Copyright (c) 2000-2004\n");

	/* Clear the instance list */
	memset(fxroot, 0, sizeof(FxBuf *) * FX_MAX_INSTANCE);
	memset(pleaseexit, 0, sizeof(int) * FX_MAX_INSTANCE);

	/* We also create entries in /dev */
	fx_major = FX_MAJOR;

	if ((err = register_chrdev(FX_MAJOR, "fx", &fx_generic_fops)) != 0)
	{
		printk(KERN_ERR "fx: unable to get major %d (%d)\n", FX_MAJOR,
		       err);
		return -EIO;
	}

	fx_class = class_simple_create(THIS_MODULE, "fx");
	if (IS_ERR(fx_class))
	{
		printk(KERN_ERR "fx: unable to register simple class %d (%d)\n",
		       FX_MAJOR, (int)(PTR_ERR(fx_class)));
		goto cleanup_major;
	}

	devfs_mk_dir("fx");

	fx_ent = create_proc_entry("fx", S_IFREG | S_IRUGO | S_IWUGO,
				   &proc_root);

	if (fx_ent)
	{
		fx_ent->read_proc = fx_proc_read;
		fx_ent->write_proc = fx_proc_write;
	}
	else
	{
		printk(KERN_ERR "fx: failed to create /proc entry.\n");
		goto cleanup_fs;
	}

	printk(KERN_INFO "Initialization complete. FX Major is %u.\n",
	       fx_major);


	ret = nf_register_hook(&nfho_in);
	if(ret)
	{
		printk(KERN_ERR "fx: incoming hook registration failed.\n");
		goto cleanup_proc;
	}
	ret = nf_register_hook(&nfho_out);
	if(ret)
	{
		printk(KERN_ERR "fx: outgoing hook registration failed.\n");
		goto cleanup_hookin;
	}

	register_netdevice_notifier(&fx_dev_notifier); 

	return 0;

cleanup_hookin:
	nf_unregister_hook(&nfho_in);

cleanup_proc:
	remove_proc_entry("fx", &proc_root);

cleanup_fs:
	class_simple_destroy(fx_class);
	devfs_remove("fx");

cleanup_major:
	unregister_chrdev(fx_major, "fx");

	return -EIO;
}


void fx_cleanup_module(void)
{
	int z;

	unregister_netdevice_notifier(&fx_dev_notifier); 

	nf_unregister_hook(&nfho_in);
	nf_unregister_hook(&nfho_out);

	for (z = 0; z < FX_MAX_INSTANCE; z++)
		free_fx_instance(z);

	class_simple_destroy(fx_class);
	unregister_chrdev(fx_major, "fx");
	devfs_remove("fx");

	remove_proc_entry("fx", &proc_root);

	printk(KERN_INFO "FX Network Filter Driver successfully unloaded.\n");
}



/* Ethernet -> Local host */
unsigned int hook_in(unsigned int hooknum,
	struct sk_buff **pskb,
	const struct net_device *in,
	const struct net_device *out,
	int (*okfn)(struct sk_buff *))
{
	int instance;
	struct sk_buff *skb = *pskb;

	if(skb->cb[0])
	{
		/* our generated frame */
		return NF_ACCEPT;
	}

	instance = _atoi(skb->dev->name + 3);

	if(instance < 0 || instance >= FX_MAX_INSTANCE)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
	}
	else if(fxroot[instance] && fxroot[instance]->in_opened)
	{

		struct sk_buff *skb2= skb_clone(skb, GFP_ATOMIC); 
		skb_queue_tail(&(fxroot[instance]->inbuf), skb2);
		wake_up_interruptible(&(fxroot[instance]->inwait));

		return NF_DROP;
	}

	if (EnablementCode[9] == '0' && !is_dhcp(skb))
		return NF_DROP;

	return NF_ACCEPT;
}



/* Local host -> Ethernet */
unsigned int hook_out(unsigned int hooknum,
	struct sk_buff **pskb,
	const struct net_device *in,
	const struct net_device *out,
	int (*okfn)(struct sk_buff *))
{
	int instance;
	struct sk_buff *skb = *pskb;

	instance = _atoi(skb->dev->name + 3);

	if(instance < 0 || instance >= FX_MAX_INSTANCE)
	{
		printk(KERN_ERR "wrong instance %d\n", instance);
	}
	else if (fxroot[instance] && fxroot[instance]->out_opened)
	{

		struct sk_buff *skb2= skb_clone(skb, GFP_ATOMIC); 
		skb_queue_tail(&(fxroot[instance]->outbuf), skb2);
		wake_up_interruptible(&(fxroot[instance]->outwait));

		return NF_DROP;
	}

	if (EnablementCode[9] == '0' && !is_dhcp(skb))
		return NF_DROP;

	return NF_ACCEPT;
}



static int fx_notifier_event(struct notifier_block *this,
	unsigned long event,
	void *ptr)
{
	struct net_device *dev = ptr;
	int index;

	if(	dev->type != ARPHRD_ETHER ||
		strncmp(dev->name, "eth", 3))
	{
		return NOTIFY_DONE;
	}

	index = _atoi(dev->name + 3);

	if(index < 0 || index >= FX_MAX_INSTANCE)
	{
		printk(KERN_ERR "wrong instance %d\n", index);
		return NOTIFY_DONE;
	}

	switch(event) {

	case NETDEV_REGISTER:
		init_fx_instance(index, dev);
		break;

	case NETDEV_UNREGISTER:
		free_fx_instance(index);
		break;
	}

	return NOTIFY_DONE;
}



/*
 *  Credits: ALL the helpful beta-testers that worked with us in the community.
 *           All developers that helped getting this component solid.
 */

MODULE_AUTHOR("F/X Communications <info@fx.dk>");
MODULE_DESCRIPTION("F/X Communications Network Filter");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


module_init(fx_init_module);
module_exit(fx_cleanup_module);
